from unittest import mock

import pytest

from autothalix.measurements import LinearSweepVoltammetry

@pytest.fixture
def lsv_measurement(mocker):
    """Create a mock instance of LinearSweepVoltammetry with mocked `wr_connection`"""
    # create a mock instance of LinearSweepVoltammetry with mocked `wr_connection`
    wr_connection = mocker.MagicMock()
    lsv_measurement = LinearSweepVoltammetry(wr_connection, 'test_lsv')
    return lsv_measurement


def test_param_types(lsv_measurement):
    """Test that the parameters are of the correct type."""
    assert isinstance(lsv_measurement.absolute_tolerance, float)
    assert isinstance(lsv_measurement.counter, int)
    assert isinstance(lsv_measurement.first_edge_potential, float)
    assert isinstance(lsv_measurement.first_edge_potential_relation, str)
    assert isinstance(lsv_measurement.fourth_edge_potential, float)
    assert isinstance(lsv_measurement.fourth_edge_potential_relation, str)
    assert isinstance(lsv_measurement.maximum_current, float)
    assert isinstance(lsv_measurement.maximum_waiting_time, float)
    assert isinstance(lsv_measurement.minimum_current, float)
    assert isinstance(lsv_measurement.minimum_waiting_time, float)
    assert isinstance(lsv_measurement.naming, str)
    assert isinstance(lsv_measurement.ohmic_drop, float)
    assert isinstance(lsv_measurement.output_path, str)
    assert isinstance(lsv_measurement.potential_resolution, float)
    assert isinstance(lsv_measurement.relative_tolerance, float)
    assert isinstance(lsv_measurement.scan_rate, float)
    assert isinstance(lsv_measurement.second_edge_potential, float)
    assert isinstance(lsv_measurement.second_edge_potential_relation, str)
    assert isinstance(lsv_measurement.sweep_mode, str)
    assert isinstance(lsv_measurement.third_edge_potential, float)
    assert isinstance(lsv_measurement.third_edge_potential_relation, str)

def test_measurement_name(lsv_measurement):
    """Test that the measurement name is correct."""
    assert lsv_measurement.measurement_name == 'lsv'


def test_parameters(lsv_measurement):
    """Test that the parameters are correct."""
    assert set(lsv_measurement.parameters) == {
            'absolute_tolerance',
            'counter',
            'first_edge_potential',
            'first_edge_potential_relation',
            'fourth_edge_potential',
            'fourth_edge_potential_relation',
            'maximum_current',
            'maximum_waiting_time',
            'minimum_current',
            'minimum_waiting_time',
            'naming',
            'ohmic_drop',
            'output_path',
            'potential_resolution',
            'relative_tolerance',
            'scan_rate',
            'second_edge_potential',
            'second_edge_potential_relation',
            'sweep_mode',
            'third_edge_potential',
            'third_edge_potential_relation',
    }


def test_send_parameters(lsv_measurement):
    """Test that the parameters are sent correctly."""
    lsv_measurement._send_parameters()
    lsv_measurement.wr_connection.setIEOutputPath.assert_called_once_with(lsv_measurement.output_path)
    lsv_measurement.wr_connection.setIEOutputFileName.assert_called_once_with(lsv_measurement._output_filename)
    lsv_measurement.wr_connection.setIENaming.assert_called_once_with(lsv_measurement.naming)
    lsv_measurement.wr_connection.setIEScanRate.assert_called_once_with(lsv_measurement.scan_rate)
    lsv_measurement.wr_connection.setIEPotentialResolution.assert_called_once_with(lsv_measurement.potential_resolution)
    lsv_measurement.wr_connection.setIEFirstEdgePotential.assert_called_once_with(lsv_measurement.first_edge_potential)
    lsv_measurement.wr_connection.setIEFirstEdgePotentialRelation.assert_called_once_with(lsv_measurement.first_edge_potential_relation)
    lsv_measurement.wr_connection.setIESecondEdgePotential.assert_called_once_with(lsv_measurement.second_edge_potential)
    lsv_measurement.wr_connection.setIESecondEdgePotentialRelation.assert_called_once_with(lsv_measurement.second_edge_potential_relation)
    lsv_measurement.wr_connection.setIEThirdEdgePotential.assert_called_once_with(lsv_measurement.third_edge_potential)
    lsv_measurement.wr_connection.setIEThirdEdgePotentialRelation.assert_called_once_with(lsv_measurement.third_edge_potential_relation)
    lsv_measurement.wr_connection.setIEFourthEdgePotential.assert_called_once_with(lsv_measurement.fourth_edge_potential)
    lsv_measurement.wr_connection.setIEFourthEdgePotentialRelation.assert_called_once_with(lsv_measurement.fourth_edge_potential_relation)
    lsv_measurement.wr_connection.setIESweepMode.assert_called_once_with(lsv_measurement.sweep_mode)
    lsv_measurement.wr_connection.setIEMinimumCurrent.assert_called_once_with(lsv_measurement.minimum_current)
    lsv_measurement.wr_connection.setIEMaximumCurrent.assert_called_once_with(lsv_measurement.maximum_current)
    lsv_measurement.wr_connection.setIEMinimumWaitingTime.assert_called_once_with(lsv_measurement.minimum_waiting_time)
    lsv_measurement.wr_connection.setIEMaximumWaitingTime.assert_called_once_with(lsv_measurement.maximum_waiting_time)
    lsv_measurement.wr_connection.setIEAbsoluteTolerance.assert_called_once_with(lsv_measurement.absolute_tolerance)
    lsv_measurement.wr_connection.setIERelativeTolerance.assert_called_once_with(lsv_measurement.relative_tolerance)
    lsv_measurement.wr_connection.setIEOhmicDrop.assert_called_once_with(lsv_measurement.ohmic_drop)



def test_run(lsv_measurement):
    """Test that the measurement is run correctly."""
    lsv_measurement.run()
    lsv_measurement.wr_connection.measureIE.assert_called_once_with()


import pytest
from unittest import mock

from autothalix.measurements import ElectrochemicalImpedanceSpectroscopy

@pytest.fixture
def eis_measurement(mocker):
    """Create a mock instance of ElectrochemicalImpedanceSpectroscopy with mocked `wr_connection`"""
    # create a mock instance of ElectrochemicalImpedanceSpectroscopy with mocked `wr_connection`
    wr_connection = mocker.MagicMock()
    eis_measurement = ElectrochemicalImpedanceSpectroscopy(wr_connection, 'test_eis')
    return eis_measurement


def test_param_types(eis_measurement):
    """Test that the parameters are of the correct type."""
    assert isinstance(eis_measurement.amplitude, float)
    assert isinstance(eis_measurement.lower_frequency_limit, float)
    assert isinstance(eis_measurement.lower_number_of_periods, int)
    assert isinstance(eis_measurement.lower_steps_per_decade, float)
    assert isinstance(eis_measurement.naming, str)
    assert isinstance(eis_measurement.output_path, str)
    assert isinstance(eis_measurement.potential, float)
    # assert isinstance(eis_measurement.potentiostat_mode, int)
    assert isinstance(eis_measurement.scan_direction, str)
    assert isinstance(eis_measurement.scan_strategy, str)
    assert isinstance(eis_measurement.start_frequency, float)
    assert isinstance(eis_measurement.upper_frequency_limit, float)
    assert isinstance(eis_measurement.upper_number_of_periods, int)
    assert isinstance(eis_measurement.upper_steps_per_decade, float)


def test_measurement_name(eis_measurement):
    """Test if the measurement name is correct"""
    assert eis_measurement.measurement_name == 'eis'


def test_parameters(eis_measurement):
    """Test if the parameters are correct"""
    assert set(eis_measurement.parameters) == {
        'amplitude',
        'lower_frequency_limit',
        'lower_number_of_periods',
        'lower_steps_per_decade',
        'naming',
        'output_path',
        'potentiostat_mode',
        'potential',
        'scan_direction',
        'scan_strategy',
        'start_frequency',
        'upper_frequency_limit',
        'upper_number_of_periods',
        'upper_steps_per_decade',
    }


def test_send_parameters(eis_measurement):
    """Test if the parameters are sent correctly"""
    eis_measurement._send_parameters()
    eis_measurement.wr_connection.setPotentiostatMode.assert_called_once_with(eis_measurement.potentiostat_mode)
    eis_measurement.wr_connection.setScanStrategy.assert_called_once_with(eis_measurement.scan_strategy)
    eis_measurement.wr_connection.setScanDirection.assert_called_once_with(eis_measurement.scan_direction)
    eis_measurement.wr_connection.setAmplitude.assert_called_once_with(eis_measurement.amplitude)
    eis_measurement.wr_connection.setStartFrequency.assert_called_once_with(eis_measurement.start_frequency)
    eis_measurement.wr_connection.setLowerFrequencyLimit.assert_called_once_with(eis_measurement.lower_frequency_limit)
    eis_measurement.wr_connection.setUpperFrequencyLimit.assert_called_once_with(eis_measurement.upper_frequency_limit)
    eis_measurement.wr_connection.setLowerStepsPerDecade.assert_called_once_with(eis_measurement.lower_steps_per_decade)
    eis_measurement.wr_connection.setUpperStepsPerDecade.assert_called_once_with(eis_measurement.upper_steps_per_decade)
    eis_measurement.wr_connection.setLowerNumberOfPeriods.assert_called_once_with(eis_measurement.lower_number_of_periods)
    eis_measurement.wr_connection.setUpperNumberOfPeriods.assert_called_once_with(eis_measurement.upper_number_of_periods)
    eis_measurement.wr_connection.setPotential.assert_called_once_with(eis_measurement.potential)
    eis_measurement.wr_connection.setEISNaming.assert_called_once_with(eis_measurement.naming)
    eis_measurement.wr_connection.setEISOutputPath.assert_called_once_with(eis_measurement.output_path)

def test_run(eis_measurement):
    """Test that the measurement is run correctly."""
    eis_measurement.run()
    eis_measurement.wr_connection.measureEIS.assert_called_once_with()



import os
import time
from abc import ABC, abstractmethod
from datetime import datetime

import yaml
from thales_remote.script_wrapper import PotentiostatMode
from thales_remote.script_wrapper import ThalesRemoteScriptWrapper

from autothalix.logging import logger
from autothalix.utils import write_dict_to_csv, safe_pot


class BaseMeasurement(ABC):
    def __init__(self, wr_connection: ThalesRemoteScriptWrapper, measurement_id: str, **kwargs):
        """
        :param wr_connection: ThalesRemoteScriptWrapper object to communicate with Thales. You can create it with
            initialize_experiment() function. See autothalix.utils
        :param kwargs: For required parameters look into attribute "parameters" or into baseline file
        :param measurement_id: Unique identifier of the measurement. It will be used in filename with results
        """
        self.load_baseline()  # sets default parameters for a measurement
        self.current_datetime = datetime.today().strftime('%d_%m_%Y_%H_%M_%S')
        self.measurement_id = measurement_id
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._check_parameters()
        self.wr_connection = wr_connection

    def _check_connection(self):
        return self.wr_connection._remote_connection.isConnectedToTerm()

    def _check_parameters(self):
        """Checks if all mandatory parameters are set"""
        for parameter in self.parameters:
            if not hasattr(self, parameter):
                raise ValueError(f'Parameter {parameter} is not set. Check if this parameters in '
                                 f'baseline file {self._baseline_path} and if it is set in the script.')

    @property
    def _baseline_path(self):
        return 'baseline.yaml'

    def load_baseline(self):
        """ Loads baseline parameters from baseline.yaml file """
        with open('baseline.yaml', 'r') as file:
            data = yaml.safe_load(file)[self.measurement_name]
        for parameter_dict in data:
            parameter_value = list(parameter_dict.values())[0]
            parameter_name = list(parameter_dict.keys())[0]
            setattr(self, parameter_name, parameter_value)

    @property
    def _output_filename(self):
        return f"{self.measurement_name}_{self.measurement_id}_{self.current_datetime}"

    @property
    def _run_message(self):
        message = f"Running {self.__str__()} with following parameters:\n"
        for parameter in self.parameters:
            message += f"\t{parameter}: {getattr(self, parameter)}\n"
        return message

    def __str__(self):
        return self._measurement_name

    @property
    def measurement_name(self):
        """
        Name of the experiment
        """
        return self._measurement_name

    def run(self):
        """
        Runs the measurement.

        First checks if connection is established, then sends parameters to the potentiostat and
        starts the measurement.

        :return: bool
            Returns True if measurement was successful.

        :raises ConnectionError:
            If connection is not established, raises ConnectionError with message
            'Connection is not established. Check that connection is established and try again.'
        """
        if self._check_connection():
            logger.info(self._run_message)
            self._send_parameters()
            self._start_measurements()
        else:
            raise ConnectionError('Connection is not established. Check that connection is established and try again.')
        return True

    @property
    @abstractmethod
    def parameters(self):
        pass

    @abstractmethod
    def _start_measurements(self):
        pass

    @abstractmethod
    def _send_parameters(self):
        pass

    @property
    def potentiostat_mode(self):
        return self._PotentiostatMode

    @potentiostat_mode.setter
    def potentiostat_mode(self, value: str):
        mapping = {
            'pseudogalvanostatic': PotentiostatMode.POTMODE_PSEUDOGALVANOSTATIC,
            'galvanostatic': PotentiostatMode.POTMODE_GALVANOSTATIC,
            'potentiostatic': PotentiostatMode.POTMODE_POTENTIOSTATIC,
        }

        self._PotentiostatMode = mapping[str.lower(value)]


class CyclicVoltammetry(BaseMeasurement):
    """
    Cyclic Voltammetry measurement class. Inherits from BaseMeasurement class."""
    _measurement_name = 'cv'
    _measurement_full_name = 'Cyclic Voltammetry'

    @property
    def parameters(self):
        """Returns a list of mandatory parameters for CV measurement"""
        return [
            'counter',
            'cycles',
            'end_hold_time',
            'end_potential',
            'lower_reversing_potential',
            'naming',
            'maximum_current',
            'minimum_current',
            'output_path',
            'ohmic_drop',
            'samples_per_cycle',
            'scan_rate',
            'start_hold_time',
            'start_potential',
            'upper_reversing_potential',
            'analog_function_generator',
            'auto_restart_at_current_overflow',
            'auto_restart_at_current_underflow'
        ]

    def _send_parameters(self):
        self.wr_connection.setCVCounter(self.counter)
        self.wr_connection.setCVOutputPath(self.output_path)
        self.wr_connection.setCVOutputFileName(self._output_filename)
        self.wr_connection.setCVNaming(self.naming)
        self.wr_connection.setCVStartPotential(self.start_potential)
        self.wr_connection.setCVUpperReversingPotential(self.upper_reversing_potential)
        self.wr_connection.setCVLowerReversingPotential(self.lower_reversing_potential)
        self.wr_connection.setCVEndPotential(self.end_potential)
        self.wr_connection.setCVStartHoldTime(self.start_hold_time)
        self.wr_connection.setCVEndHoldTime(self.end_hold_time)
        self.wr_connection.setCVCycles(self.cycles)
        self.wr_connection.setCVSamplesPerCycle(self.samples_per_cycle)
        self.wr_connection.setCVScanRate(self.scan_rate)
        self.wr_connection.setCVMaximumCurrent(self.maximum_current)
        self.wr_connection.setCVMinimumCurrent(self.minimum_current)
        self.wr_connection.setCVOhmicDrop(self.ohmic_drop)
        self.wr_connection.enableCVAnalogFunctionGenerator(self.analog_function_generator)
        self.wr_connection.enableCVAutoRestartAtCurrentOverflow(self.auto_restart_at_current_overflow)
        self.wr_connection.enableCVAutoRestartAtCurrentUnderflow(self.auto_restart_at_current_underflow)

    def _start_measurements(self):
        self.wr_connection.disableCVAutoRestartAtCurrentOverflow()
        self.wr_connection.disableCVAutoRestartAtCurrentUnderflow()
        self.wr_connection.disableCVAnalogFunctionGenerator()
        self.wr_connection.checkCVSetup()
        self.wr_connection.measureCV()


class LinearSweepVoltammetry(BaseMeasurement):
    _measurement_name = 'lsv'
    _measurement_full_name = 'Linear Sweep Voltammetry'

    @property
    def parameters(self):
        """Returns a list of mandatory parameters for LSV measurement"""
        return [
            'absolute_tolerance',
            'counter',
            'first_edge_potential',
            'first_edge_potential_relation',
            'fourth_edge_potential',
            'fourth_edge_potential_relation',
            'maximum_current',
            'maximum_waiting_time',
            'minimum_current',
            'minimum_waiting_time',
            'naming',
            'ohmic_drop',
            'output_path',
            'potential_resolution',
            'relative_tolerance',
            'scan_rate',
            'second_edge_potential',
            'second_edge_potential_relation',
            'sweep_mode',
            'third_edge_potential',
            'third_edge_potential_relation',
        ]

    def _send_parameters(self):
        self.wr_connection.setIEAbsoluteTolerance(self.absolute_tolerance)
        self.wr_connection.setIECounter(self.counter)
        self.wr_connection.setIEFirstEdgePotential(self.first_edge_potential)
        self.wr_connection.setIEFirstEdgePotentialRelation(self.first_edge_potential_relation)
        self.wr_connection.setIEFourthEdgePotential(self.fourth_edge_potential)
        self.wr_connection.setIEFourthEdgePotentialRelation(self.fourth_edge_potential_relation)
        self.wr_connection.setIEMaximumCurrent(self.maximum_current)
        self.wr_connection.setIEMaximumWaitingTime(self.maximum_waiting_time)
        self.wr_connection.setIEMinimumCurrent(self.minimum_current)
        self.wr_connection.setIEMinimumWaitingTime(self.minimum_waiting_time)
        self.wr_connection.setIENaming(self.naming)
        self.wr_connection.setIEOhmicDrop(self.ohmic_drop)
        self.wr_connection.setIEOutputFileName(self._output_filename)
        self.wr_connection.setIEOutputPath(self.output_path)
        self.wr_connection.setIEPotentialResolution(self.potential_resolution)
        self.wr_connection.setIERelativeTolerance(self.relative_tolerance)
        self.wr_connection.setIEScanRate(self.scan_rate)
        self.wr_connection.setIESecondEdgePotential(self.second_edge_potential)
        self.wr_connection.setIESecondEdgePotentialRelation(self.second_edge_potential_relation)
        self.wr_connection.setIESweepMode(self.sweep_mode)
        self.wr_connection.setIEThirdEdgePotential(self.third_edge_potential)
        self.wr_connection.setIEThirdEdgePotentialRelation(self.third_edge_potential_relation)

    def _start_measurements(self):
        self.wr_connection.checkIESetup()
        self.wr_connection.measureIE()


class ElectrochemicalImpedanceSpectroscopy(BaseMeasurement):
    """ Electrochemical Impedance Spectroscopy measurement class"""

    _measurement_name = 'eis'
    _measurement_full_name = 'Electrochemical Impedance Spectroscopy'

    def __str__(self):
        return self._measurement_name

    def _send_parameters(self):
        self.wr_connection.setPotentiostatMode(self.potentiostat_mode)
        self.wr_connection.setAmplitude(self.amplitude)
        self.wr_connection.setPotential(self.potential)
        self.wr_connection.setLowerFrequencyLimit(self.lower_frequency_limit)
        self.wr_connection.setStartFrequency(self.start_frequency)
        self.wr_connection.setUpperFrequencyLimit(self.upper_frequency_limit)
        self.wr_connection.setLowerNumberOfPeriods(self.lower_number_of_periods)
        self.wr_connection.setLowerStepsPerDecade(self.lower_steps_per_decade)
        self.wr_connection.setUpperNumberOfPeriods(self.upper_number_of_periods)
        self.wr_connection.setUpperStepsPerDecade(self.upper_steps_per_decade)
        self.wr_connection.setScanDirection(self.scan_direction)
        self.wr_connection.setScanStrategy(self.scan_strategy)
        self.wr_connection.setEISOutputPath(self.output_path)
        self.wr_connection.setEISOutputFileName(self._output_filename)
        self.wr_connection.setEISNaming(self.naming)

    @property
    def parameters(self):
        """Returns a list of mandatory parameters for IE measurement"""
        return [
            "potentiostat_mode",
            "amplitude",
            "potential",
            "lower_frequency_limit",
            "start_frequency",
            "upper_frequency_limit",
            "lower_number_of_periods",
            "lower_steps_per_decade",
            "upper_number_of_periods",
            "upper_steps_per_decade",
            "scan_direction",
            "scan_strategy",
            "output_path",
            "naming",
        ]

    def _start_measurements(self):
        self.wr_connection.enablePotentiostat()
        self.wr_connection.measureEIS()
        self.wr_connection.disablePotentiostat()


class BaseManualMeasurements(BaseMeasurement, ABC):
    """
    Base class for manual measurements
    """

    def __init__(self, wr_connection: ThalesRemoteScriptWrapper, measurement_id: str, **kwargs):
        super().__init__(wr_connection, measurement_id, **kwargs)

    def _save_data(self):
        logger.info(f"Saving {self.measurement_name} data to {self._output_filename}.csv")
        write_dict_to_csv(self.measured_data, os.path.join(self.output_path, self._output_filename + '.csv'))

    def run(self):
        super().run()
        self._save_data()


class OpenCircuitPotential(BaseManualMeasurements):
    """
    Open Circuit Potential measurement class. This class is used to measure the open circuit potential of the sample.
    It is a manually implemented measurement. For more information on implementation please refer _start_measurements
    method.
    """
    _measurement_name = 'ocp'

    def __str__(self):
        return 'Open Circuit Potential'

    @property
    def parameters(self):
        """Returns a list of mandatory parameters for OCP measurement"""
        return [
            'current',
            'delta',
            'output_path',
            'potentiostat_mode',
            'seconds',
        ]

    def _send_parameters(self):
        self.wr_connection.setPotentiostatMode(self.potentiostat_mode)
        self.wr_connection.setCurrent(self.current)

    @safe_pot
    def _start_measurements(self):
        measured_data = {'time': [], 'potential_V': []}
        for i in range(0, self.seconds, self.delta):
            potential = self.wr_connection.getPotential()
            logger.info(f'Second:\t{i}\tPotential:\t{potential}V')
            measured_data['time'].append(i)
            measured_data['potential_V'].append(potential)
            time.sleep(self.delta)
        self.measured_data = measured_data
        return True


class Impedance(BaseManualMeasurements):
    """
    This class is used to measure impedance of a cell. It is a manually implemented measurement. For more information
    on implementation please refer _start_measurements method.
    """
    _measurement_name = 'imp'

    def __str__(self):
        return 'Impedance'

    @property
    def parameters(self):
        return [
            'amplitude',
            'current',
            'delta',
            'frequency',
            'number_of_periods',
            'output_path',
            'potentiostat_mode',
            'seconds',
        ]

    def _send_parameters(self):
        self.wr_connection.setPotentiostatMode(self.potentiostat_mode)
        self.wr_connection.setCurrent(self.current)

    @safe_pot
    def _start_measurements(self):
        """
        Start measurements for Impedance measurement
        """
        measured_data = {'time': [], 'impedance_Ohm': [], 'phase_deg': []}
        for i in range(0, self.seconds, self.delta):
            response = self.wr_connection.getImpedance(
                frequency=self.frequency,
                amplitude=self.amplitude,
                number_of_periods=self.number_of_periods)

            imp = float(response.real)
            phase = float(response.imag)
            measured_data['time'].append(i)
            measured_data['impedance_Ohm'].append(imp)
            measured_data['phase_deg'].append(phase)
            logger.info(f'Seconds:\t{i}\tImpedance:\t {imp} Ohm\tPhase:\t{phase}°')
            time.sleep(self.delta)
        self.measured_data = measured_data
        return True


class ChronoAmperometry(BaseManualMeasurements):
    """
    This class will perform a chronoamperometry measurement.
    """
    _measurement_name = 'ca'

    def __str__(self):
        return 'Chronoamperometry'

    @property
    def parameters(self):
        return [
            'induction_pot',
            'induction_t',
            'electrolysis_pot',
            'electrolysis_t',
            'relaxation_pot',
            'relaxation_t',
            'sample_rate',
            'potentiostat_mode',
            'output_path',
        ]

    def _send_parameters(self):
        self.wr_connection.setPotentiostatMode(self.potentiostat_mode)

    def _set_induction(self):
        self.wr_connection.setPotential(self.induction_pot)

    def _set_electrolysis(self):
        self.wr_connection.setPotential(self.electrolysis_pot)

    def _set_relaxation(self):
        self.wr_connection.setPotential(self.relaxation_pot)

    @safe_pot
    def _start_measurements(self):
        measured_data = {'time': [], 'current_A': []}
        # induction phase
        logger.info(f'Induction phase for {self.induction_t} seconds with {self.induction_pot} V')
        self._set_induction()
        time.sleep(self.induction_t)

        # electrolysis phase
        logger.info(f'Electrolysis phase for {self.electrolysis_t} seconds with {self.electrolysis_pot} V')
        start_time = datetime.now().timestamp()  # get start phase time
        self._set_electrolysis()
        while (datetime.now().timestamp() - start_time) < self.electrolysis_t:
            current = self.wr_connection.getCurrent()
            seconds = datetime.now().timestamp() - start_time
            measured_data['time'].append(seconds)
            measured_data['current_A'].append(current)
            logger.info(f'Seconds:\t{seconds}\tCurrent:\t {current} A')
            time.sleep(1 / self.sample_rate)  # if sample rate 0.5 will wait 2 seconds between samples

        # relaxation phase
        logger.info(f'Relaxation phase for {self.relaxation_t} seconds with {self.relaxation_pot} V')
        self._set_relaxation()
        time.sleep(self.relaxation_t)

        self.measured_data = measured_data



import logging
import os
import sys
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


script_path = sys.argv[0]
formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
file_handler = logging.FileHandler(os.path.join(os.path.dirname(script_path), 'experiment.log'))
file_handler.setFormatter(formatter)
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)
logger.addHandler(file_handler)
logger.addHandler(console_handler)


import pytest

from autothalix.measurements import CyclicVoltammetry


@pytest.fixture
def cv_measurement(mocker):
    """Create a mock instance of CyclicVoltammetry with mocked `wr_connection`"""
    wr_connection = mocker.MagicMock() # wr_connection is a mock
    cv_measurement = CyclicVoltammetry(wr_connection, 'test_cv') # create a mock instance of CyclicVoltammetry
    return cv_measurement


def test_param_types(cv_measurement):
    """Test that the parameters are of the correct type."""
    assert type(cv_measurement.counter) == int
    assert type(cv_measurement.cycles) == float
    assert type(cv_measurement.end_hold_time) == float
    assert type(cv_measurement.end_potential) == float
    assert type(cv_measurement.lower_reversing_potential) == float
    assert type(cv_measurement.naming) == str
    assert type(cv_measurement.maximum_current) == float
    assert type(cv_measurement.minimum_current) == float
    assert type(cv_measurement.output_path) == str
    assert type(cv_measurement.ohmic_drop) == float
    assert type(cv_measurement.samples_per_cycle) == int
    assert type(cv_measurement.scan_rate) == float
    assert type(cv_measurement.start_hold_time) == float
    assert type(cv_measurement.start_potential) == float
    assert type(cv_measurement.upper_reversing_potential) == float

def test_measurement_name(cv_measurement):
    """Test that the measurement name is correct."""
    assert cv_measurement.measurement_name == 'cv'


def test_parameters(cv_measurement):
    """Test that the parameters are correct."""
    assert set(cv_measurement.parameters) == {'counter', 'cycles', 'end_hold_time', 'end_potential',
                                              'lower_reversing_potential', 'naming', 'maximum_current',
                                              'minimum_current', 'output_path', 'ohmic_drop', 'samples_per_cycle',
                                              'scan_rate', 'start_hold_time', 'start_potential',
                                              'upper_reversing_potential', 'analog_function_generator',
                                              'auto_restart_at_current_overflow', 'auto_restart_at_current_underflow'}


def test_send_parameters(cv_measurement):
    """Test that the parameters are sent correctly."""
    cv_measurement._send_parameters()
    cv_measurement.wr_connection.setCVCounter.assert_called_once_with(cv_measurement.counter)
    cv_measurement.wr_connection.setCVOutputPath.assert_called_once_with(cv_measurement.output_path)
    cv_measurement.wr_connection.setCVOutputFileName.assert_called_once_with(cv_measurement._output_filename)
    cv_measurement.wr_connection.setCVNaming.assert_called_once_with(cv_measurement.naming)
    cv_measurement.wr_connection.setCVStartPotential.assert_called_once_with(cv_measurement.start_potential)
    cv_measurement.wr_connection.setCVUpperReversingPotential.assert_called_once_with(
        cv_measurement.upper_reversing_potential)
    cv_measurement.wr_connection.setCVLowerReversingPotential.assert_called_once_with(
        cv_measurement.lower_reversing_potential)
    cv_measurement.wr_connection.setCVEndPotential.assert_called_once_with(cv_measurement.end_potential)
    cv_measurement.wr_connection.setCVStartHoldTime.assert_called_once_with(cv_measurement.start_hold_time)
    cv_measurement.wr_connection.setCVEndHoldTime.assert_called_once_with(cv_measurement.end_hold_time)
    cv_measurement.wr_connection.setCVCycles.assert_called_once_with(cv_measurement.cycles)
    cv_measurement.wr_connection.setCVSamplesPerCycle.assert_called_once_with(cv_measurement.samples_per_cycle)
    cv_measurement.wr_connection.setCVScanRate.assert_called_once_with(cv_measurement.scan_rate)
    cv_measurement.wr_connection.setCVMaximumCurrent.assert_called_once_with(cv_measurement.maximum_current)
    cv_measurement.wr_connection.setCVMinimumCurrent.assert_called_once_with(cv_measurement.minimum_current)
    cv_measurement.wr_connection.setCVOhmicDrop.assert_called_once_with(cv_measurement.ohmic_drop)


def test_start_measurements(cv_measurement):
    """Test that the measurements are started correctly."""
    cv_measurement._start_measurements()
    cv_measurement.wr_connection.disableCVAutoRestartAtCurrentOverflow.assert_called_once()
    cv_measurement.wr_connection.disableCVAutoRestartAtCurrentUnderflow.assert_called_once()
    cv_measurement.wr_connection.disableCVAnalogFunctionGenerator.assert_called_once()
    cv_measurement.wr_connection.checkCVSetup.assert_called_once()
    cv_measurement.wr_connection.measureCV.assert_called_once()



import csv
from autothalix.logging import logger
from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import ThalesRemoteScriptWrapper


def initialize_experiment():
    """
    Initialize the experiment by connecting to the Thales Zennium and calibrating the offsets
    :return: zennium_connection, zahner_zennium
    :rtype: ThalesRemoteConnection, ThalesRemoteScriptWrapper
    """
    zennium_connection = ThalesRemoteConnection()
    zennium_connection.connectToTerm("localhost", "ScriptRemote")  # do not change to anything besides localhost
    zahner_zennium = ThalesRemoteScriptWrapper(zennium_connection)
    zahner_zennium.forceThalesIntoRemoteScript()
    zahner_zennium.calibrateOffsets()
    return zennium_connection, zahner_zennium


def write_dict_to_csv(dict_data, file_path):
    """
    Write a dictionary to a csv file with the keys as the header row and the values as the data rows
    :param dict_data:
    :param file_path:
    :return:
    """
    with open(file_path, mode='w', newline='') as file:
        writer = csv.writer(file)

        # Write the header row
        writer.writerow(dict_data.keys())

        # Write the data rows
        for row in zip(*dict_data.values()):
            writer.writerow(row)


def safe_pot(func):
    def wrapper(self, *args, **kwargs):
        logger.info('Manually enabling potentiostat')
        self.wr_connection.enablePotentiostat()
        try:
            result = func(self, *args, **kwargs)
        except Exception as e:  # disables potentiostat if an error occurs
            self.wr_connection.disablePotentiostat()
            logger.error(f'Error occurred during {str(self)} measurements: "{e}"')
            logger.info("Potentiostat disabled")
            raise e
        self.wr_connection.disablePotentiostat()
        logger.info("Potentiostat disabled")
        return result

    return wrapper



from unittest import mock

import pytest

from autothalix.measurements import OpenCircuitPotential


@pytest.fixture
def ocp_measurement(mocker: mock):
    """Create a mock instance of OpenCircuitPotential with mocked `wr_connection`"""
    wr_connection = mocker.MagicMock()
    wr_connection.getPotential.return_value = 1.1
    ocp_measurement = OpenCircuitPotential(wr_connection, 'test_ocp')
    return ocp_measurement


def test_measurement_name(ocp_measurement):
    """Test the measurement name"""
    assert ocp_measurement.measurement_name == 'ocp'


def test_parameters(ocp_measurement):
    """Test the parameters"""
    assert ocp_measurement.parameters == [
        'current',
        'delta',
        'output_path',
        'potentiostat_mode',
        'seconds',
    ]


def test_send_parameters(ocp_measurement):
    """Test the parameters are sent to the potentiostat"""
    ocp_measurement._send_parameters()
    ocp_measurement.wr_connection.setPotentiostatMode.assert_called_once_with(ocp_measurement.potentiostat_mode)
    ocp_measurement.wr_connection.setCurrent.assert_called_once_with(ocp_measurement.current)


def test_start_measurements_diff_delta(ocp_measurement):
    """Test the measurements are taken at the correct intervals"""
    ocp_measurement.seconds = 4
    ocp_measurement.delta = 2  # 2 seconds
    print(ocp_measurement.seconds, print(ocp_measurement.delta))
    ocp_measurement._start_measurements()
    ocp_measurement.wr_connection.getPotential.call_count == 2


def test_start_measurements(ocp_measurement):
    """Test the measurements are taken at the correct intervals"""
    ocp_measurement.seconds = 6
    print(ocp_measurement.seconds, print(ocp_measurement.delta))
    ocp_measurement._start_measurements()
    ocp_measurement.wr_connection.enablePotentiostat.assert_called_once()
    ocp_measurement.wr_connection.disablePotentiostat.assert_called_once()
    ocp_measurement.wr_connection.getPotential.call_count == 6


def test_start_measurements_results_2_sec(ocp_measurement):
    """Test the process of the measurements"""
    ocp_measurement.seconds = 4
    ocp_measurement.delta = 2  # 2 seconds
    ocp_measurement._start_measurements()
    print(ocp_measurement.seconds, print(ocp_measurement.delta))
    ocp_measurement._start_measurements()
    assert len(ocp_measurement.measured_data) == 2
    for i in range(2):
        assert ocp_measurement.measured_data["potential_V"][i] == 1.1
        assert ocp_measurement.measured_data["time"][i] == i * 2


def test_start_measurements_results_1_sec(ocp_measurement):
    """Test the process of the measurements"""
    ocp_measurement.seconds = 4
    ocp_measurement.delta = 1  # 2 seconds
    ocp_measurement._start_measurements()
    print(ocp_measurement.seconds, print(ocp_measurement.delta))
    ocp_measurement._start_measurements()
    assert len(ocp_measurement.measured_data["potential_V"]) == 4
    assert len(ocp_measurement.measured_data["time"]) == 4
    for i in range(0, 4):
        assert ocp_measurement.measured_data["potential_V"][i] == 1.1
        assert ocp_measurement.measured_data["time"][i] == i



from unittest import mock

import pytest

from autothalix.measurements import ChronoAmperometry


@pytest.fixture
def ca_measurement(mocker: mock):
    """Create a mock instance of OpenCircuitPotential with mocked `wr_connection`"""
    wr_connection = mocker.MagicMock()
    wr_connection.getPotential.return_value = 1.1
    ca_measurement = ChronoAmperometry(wr_connection, 'test_ocp')
    return ca_measurement


def test_measurement_name(ca_measurement):
    """Test the measurement name"""
    assert ca_measurement.measurement_name == 'ca'


def test_parameters(ca_measurement):
    """Test the parameters"""
    assert ca_measurement.parameters ==  [
            'induction_pot',
            'induction_t',
            'electrolysis_pot',
            'electrolysis_t',
            'relaxation_pot',
            'relaxation_t',
            'sample_rate',
            'potentiostat_mode',
            'output_path',
        ]


def test_send_parameters(ca_measurement):
    """Test the parameters are sent to the potentiostat"""
    ca_measurement._send_parameters()
    ca_measurement.wr_connection.setPotentiostatMode.assert_called_once_with(ca_measurement.potentiostat_mode)


def test_phases(ca_measurement):
    """Test that every phase called"""
    ca_measurement._start_measurements()
    ca_measurement.wr_connection.enablePotentiostat.assert_called_once()
    ca_measurement.wr_connection.disablePotentiostat.assert_called_once()
    ca_measurement.wr_connection.getPotential.call_count == 3 # 3 phases = 3 calls




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = f.readlines()

install_requires = [req.strip() for req in requirements]

setuptools.setup(
    name="AutoThaliX",
    version="0.3.1",
    author="ArkadyA",
    author_email="arkadymirz@gmail.com",
    description="A set of tools to work with tales potentiostats. Uses tales_remote",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Arkady-A/autothalix",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    install_requires=install_requires,
)



from unittest import mock

import pytest
from thales_remote.script_wrapper import ThalesRemoteScriptWrapper

from autothalix.measurements import Impedance


@pytest.fixture
def imp_measurement(mocker: mock):
    """Create a mock instance of Impedance with mocked `wr_connection`"""
    # create a mock instance of OpenCircuitPotential with mocked `wr_connection`
    wr_connection = mocker.MagicMock()  # type: ThalesRemoteScriptWrapper
    wr_connection.getImpedance.return_value = 1.1
    imp_measurement = Impedance(wr_connection, 'test_ocp')
    return imp_measurement


def test_measurement_name(imp_measurement):
    """Test if the measurement name is correct"""
    assert imp_measurement.measurement_name == 'imp'


def test_parameters(imp_measurement):
    """Test if the parameters are correct"""
    assert imp_measurement.parameters == [
        'amplitude',
        'current',
        'delta',
        'frequency',
        'number_of_periods',
        'output_path',
        'potentiostat_mode',
        'seconds',
    ]


def test_send_parameters(imp_measurement):
    """Test if the parameters are sent correctly"""
    imp_measurement._send_parameters()
    imp_measurement.wr_connection.setPotentiostatMode.assert_called_once_with(imp_measurement.potentiostat_mode)
    imp_measurement.wr_connection.setCurrent.assert_called_once_with(imp_measurement.current)


def test_start_measurements_diff_delta(imp_measurement):
    """Test if the measurements are started correctly"""
    imp_measurement.seconds = 4
    imp_measurement.delta = 2  # 2 seconds
    imp_measurement._start_measurements()
    imp_measurement.wr_connection.getImpedance.call_count == 2


def test_start_measurements_diff(imp_measurement):
    """Test if the measurements are started correctly"""
    imp_measurement.seconds = 4
    imp_measurement.delta = 1  # 1 seconds
    imp_measurement._start_measurements()
    imp_measurement.wr_connection.getImpedance.call_count == 4


def test_start_measurements_results_2_sec(imp_measurement):
    """Test if the process of the measurements is correct"""
    imp_measurement.seconds = 4
    imp_measurement.delta = 2  # 2 seconds
    imp_measurement._start_measurements()
    imp_measurement.wr_connection.getImpedance.return_value = complex(3.9, 23)
    imp_measurement._start_measurements()
    assert len(imp_measurement.measured_data["impedance_Ohm"]) == 2
    assert len(imp_measurement.measured_data["phase_deg"]) == 2
    assert len(imp_measurement.measured_data["time"]) == 2
    for i in range(0, 2):
        assert imp_measurement.measured_data["impedance_Ohm"][i] == 3.9
        assert imp_measurement.measured_data["phase_deg"][i] == 23
        assert imp_measurement.measured_data["time"][i] == i * 2


def test_start_measurements_results_1_sec(imp_measurement):
    """Test if the process of the measurements is correct"""
    imp_measurement.seconds = 4
    imp_measurement.delta = 1  # 2 seconds
    imp_measurement.wr_connection.getImpedance.return_value = complex(2.1, 20)
    imp_measurement._start_measurements()
    assert len(imp_measurement.measured_data["impedance_Ohm"]) == 4
    assert len(imp_measurement.measured_data["phase_deg"]) == 4
    assert len(imp_measurement.measured_data["time"]) == 4
    for i in range(0, 4):
        assert imp_measurement.measured_data["impedance_Ohm"][i] == 2.1
        assert imp_measurement.measured_data["phase_deg"][i] == 20
        assert imp_measurement.measured_data["time"][i] == i
